

class BaserowOrmException(Exception):
  pass


class NoRowReturned(BaserowOrmException):
  """
  This exception is raised by the #Query class if at least one row was expected but none was returned
  by the Baserow API.
  """




import copy
import dataclasses
import typing as t

from .column import Column


class Model:
  """
  Base class to represent a Baserow table. Static class members on the base class that are instances of the
  #Column class are recognized as fields that are associated with fields in a Baserow database table.
  """

  id: t.Optional[int]
  __id__: t.ClassVar[str]
  __tablename__: t.ClassVar[t.Optional[str]]
  __columns__: t.ClassVar[t.Dict[str, Column]]

  def __init_subclass__(cls) -> None:
    cls.__columns__ = {}
    if '__tablename__' not in vars(cls):
      cls.__tablename__ = None

    if not hasattr(cls, '__id__'):
      cls.__id__ = cls.__module__ + '.' + cls.__name__

    for key, value in vars(cls).items():
      if isinstance(value, Column):
        cls.__columns__[key] = value

    # Inherit parent class columns.
    for base in cls.__bases__:
      if issubclass(base, Model):
        for key, value in base.__columns__.items():
          if key not in cls.__columns__:
            cls.__columns__[key] = copy.copy(value)

    if 'id' in cls.__columns__:
      raise ValueError(f'attribute name "id" is reserved')

  def __init__(self, id: t.Optional[int] = None, **kwargs) -> None:
    self.id = id
    for key, col in self.__columns__.items():
      if key not in kwargs:
        raise TypeError(f'{type(self).__name__}(): missing keyword argument {key!r}')
      setattr(self, key, col.from_python(kwargs[key]))
    for key in kwargs:
      if key not in self.__columns__:
        raise TypeError(f'{type(self).__name__}(): unrecognized keyword argument {key!r}')

  def __repr__(self) -> str:
    primary = next(iter(self.__columns__))
    return f'{type(self).__name__}(id={self.id}, {primary}={getattr(self, primary)!r})'

  def as_dict(self) -> t.Dict[str, t.Any]:
    return {'id': self.id, **{k: getattr(self, k) for k in self.__columns__}}

  @classmethod
  def of(cls, table_name: str, field_name_overrides: t.Optional[t.Dict[str, str]] = None) -> 'ModelMappingDescription':
    return ModelMappingDescription(cls.__id__, cls.__columns__, table_name, field_name_overrides or {})


Model.__columns__ = {}


@dataclasses.dataclass
class ModelMappingDescription:
  model_id: str
  columns: t.Dict[str, Column]
  table_name: str
  field_name_overrides: t.Dict[str, str]




import typing as t
import uuid

from ..filter import Column as _Column, ValueType

if t.TYPE_CHECKING:
  from .database import Database
  from .model import Model

T_Model = t.TypeVar('T_Model', bound='Model')


class Column(_Column):
  """
  Use this class to declare a column on the class-level of a #Model subclass. Colums declared with this
  class will be assigned to a #Model instance directly (unlike, for example, the #ForeignKey column type).

  # Example

  ```py
  from baserow.orm import Column, Model

  class Product(Model):
    name = Column('Name')
    price = Column('Price')
  ```

  Columns have four types of names:

  1. The attrribute name that is assigned to them on the Model instance and that is used to reference the
     column in code.
  2. The user defined name of the column (aka. field) in the Baserow table.
  3. The Baserow internal field ID.
  4. An ORM internal UUID that is newly generated every for every Column instance. This ID is used as a
     placeholder when constructing #Filter#s which the #Query will replace with the Baserow internal
     field ID on execution.

  In the model, we only specify the attribute and user defined name. Use the #DatabaseMapping.generate()
  method, or the `baserow.orm` CLI, to generate a #DatabaseMapping which uses the user defined name to
  create a mapping from attribute name to Baserow internal field ID.
  """

  def __init__(self, user_name: str) -> None:
    """
    # Arguments
    user_name: The user specified column name in the Baserow table.
    """

    assert isinstance(user_name, str)
    # We later replace references to this ID in #Filter objects with the actual internal field ID.
    super().__init__(user_name + '.' + str(uuid.uuid4()))
    self._user_name = user_name

  def __repr__(self) -> str:
    return f'{type(self).__name__}(name={self.name!r})'

  def from_baserow(self, db: 'Database', value: t.Any) -> t.Any:
    """
    Convert a value received by Baserow for this column to the value that should be assigned to the
    Model instance attribute.
    """

    return value

  def to_baserow(self, value: t.Any) -> t.Any:
    """
    Convert a value stored on a model instance to JSON format for a request to Baserow.
    """

    return value

  def from_python(self, value: t.Any) -> t.Any:
    """
    Convert a value assigned to the row from Python.
    """

    return value

  @property
  def id(self) -> str:
    """
    The internal ID of the column that is used a as a placeholder when constructing #Filter#s.
    """

    return self._name

  @property
  def name(self) -> str:
    """
    The user defined name of the column.
    """

    return self._user_name


class ForeignKey(Column):
  """
  The #ForeignKey is a special column that should be used to represent a "link row" field in Baserow. Model
  attributes of this column type will not receive the raw data from the Baserow API as values, but instead
  an instance of #LinkedRow which automatically loads the linked rows as another Model from the database on
  access.
  """

  def __init__(self, user_name: str, model: t.Union[t.Type['Model'], t.Callable[[], t.Type['Model']]]) -> None:
    super().__init__(user_name)
    self._model = model

  def from_baserow(self, db: 'Database', value: t.Any) -> t.Any:
    refs = [Ref(x['id'], x['value']) for x in value]
    return LinkedRow(db, self.model, refs)

  def to_baserow(self, value: t.Any) -> t.Any:
    assert isinstance(value, LinkedRow)
    return [r.id for r in value.refs]

  def from_python(self, value: t.Any) -> t.Any:
    if not isinstance(value, t.Collection):
      raise TypeError(f'expected collection for column {self.name}')
    value_list = list(value)
    for item in value_list:
      if not isinstance(item, self.model):
        raise TypeError(f'expeted {self.model.__id__!r} instance for column {self.name}')
    # TODO (@NiklasRosenstein): Need to know which is the primary column in the model
    raise NotImplementedError
    # primary_col = ...
    # refs = [Ref(x.id, getattr(x, primary_col)) for x in value]
    # return LinkedRow(None, self._model, refs, {x.id: x for x in value_list})

  @property
  def model(self) -> t.Type['Model']:
    if isinstance(self._model, type):
      return self._model
    self._model = self._model()
    return self._model


class Ref(t.NamedTuple):
  """
  A reference to another row.
  """

  id: int
  name: ValueType


class LinkedRow(t.Sequence[T_Model]):
  """
  Represents a "link row" field. Loads instances of the linked Model on acces.
  """

  def __init__(
    self,
    db: t.Optional['Database'],
    model: t.Type[T_Model],
    values: t.List[Ref],
    cache: t.Optional[t.Dict[int, T_Model]] = None,
  ) -> None:
    self._db = db
    self._model = model
    self._refs = values
    self._cache: t.Dict[int, T_Model] = cache or {}

  def __repr__(self) -> str:
    return f'{type(self).__name__}({self._refs!r})'

  def __len__(self) -> int:
    return len(self._refs)

  def __iter__(self) -> t.Iterator[T_Model]:
    for i in range(len(self._refs)):
      yield self[i]

  def __getitem__(self, index: int) -> T_Model:  # type: ignore
    if index in self._cache:
      return self._cache[index]
    assert self._db is not None, 'not attached to a database'
    self._cache[index] = result = self._db._load_single(self._model, self._refs[index].id)
    return result

  @property
  def refs(self) -> t.List[Ref]:
    return self._refs




import copy
import typing as t

from ..client import BaserowClient
from ..filter import Filter, ValueType
from ..types import Page
from .column import ForeignKey
from .exc import NoRowReturned
from .mapping import DatabaseMapping
from .model import Model

T_Model = t.TypeVar('T_Model', bound='Model')


class ColumnPlaceholderTranslator(t.Dict[str, int]):
  """
  A helper class that translates #Column.id placeholders to Baserow internal field IDs.
  """

  def __init__(self, mapping: DatabaseMapping) -> None:
    self._mapping = mapping
    self._visited: t.Set[t.Type[Model]] = set()

  def visit(self, model: t.Type[Model]) -> None:
    """
    Make the runtime definition of *model* known to the converter. This is needed to ensure that we know the
    column placeholder IDs and how they are mapped to Baserow internal field IDs.
    """

    if model not in self._visited:
      for key, column in model.__columns__.items():
        self[column.id] = self._mapping.models[model.__id__].fields[key]
      self._visited.add(model)


class Database:
  """
  ORM for a Baserow database.
  """

  def __init__(self, client: BaserowClient, mapping: DatabaseMapping) -> None:
    """
    # Arguments
    client: The Baserow client to interact with.
    mapping: A previously generated mapping for the database, it's tables and internal field IDs.
    """

    self._client = client
    self._mapping = mapping
    self._translator = ColumnPlaceholderTranslator(mapping)

  def __repr__(self) -> str:
    return f'Database(db={self._mapping.database_id})'

  def _load_single(self, model: t.Type[T_Model], row_id: int) -> T_Model:
    mapping = self._mapping.models[model.__id__]
    row = self._client.get_database_table_row(mapping.table_id, row_id)
    return self._build_model_from_row(model, row)

  def _build_model_from_row(self, model: t.Type[T_Model], row: t.Dict[str, t.Any]) -> T_Model:
    mapping = self._mapping.models[model.__id__]
    row.pop('order', None)
    record_id = row.pop('id')
    record = {}
    for key, value in row.items():
      assert key.startswith('field_')
      field_id = int(key[6:])
      if field_id not in mapping.reverse_fields:
        continue
      attr_name = mapping.reverse_fields[field_id]
      column = model.__columns__[attr_name]
      record[attr_name] = column.from_baserow(self, value)
    return model(record_id, **record)

  def _preprocess_filter(self, filter: Filter) -> Filter:
    """
    Internal. Called by the #Query to replace column placeholders with Baserow internal field IDs.
    """

    if filter.field in self._translator:
      filter = copy.copy(filter)
      filter.field = f'field_{self._translator[filter.field]}'

    return filter

  def select(self, model: t.Type[T_Model]) -> 'Query[T_Model]':
    """
    Create a new #Query for rows of the given model.
    """

    self._translator.visit(model)
    return Query(self, model)

  def save(self, row: Model) -> None:
    """
    Save a model instance as a raw in its backing database. if the *row* has an ID, the method will perform an
    upadte of the row in Baserow. The #Model.id will be set after creation of a new row.
    """

    mapping = self._mapping.models[row.__id__]
    record: t.Dict[str, t.Any] = {}
    for key, col in row.__columns__.items():
      record[f'field_{mapping.fields[key]}'] = col.to_baserow(getattr(row, key))

    if row.id is None:
      row.id = self._client.create_database_table_row(mapping.table_id, record)['id']
    else:
      self._client.update_database_table_row(mapping.table_id, row.id, record)


class Query(t.Generic[T_Model]):
  """
  Represents a query for the rows of a particular model. A query can be modified until it is executed,
  after which it becomes immutable. Iterating over a query yields all matching rows as model instances.
  """

  def __init__(self, db: Database, model: t.Type[T_Model]) -> None:
    self._db = db
    self._model = model
    self._mapping = db._mapping.models[model.__id__]
    self._filters: t.List[Filter] = []
    self._page_size: t.Optional[int] = None
    self._paginator: t.Optional[t.Iterator[Page[t.Dict[str, str]]]] = None
    self._page_items: t.Optional[t.Iterator[t.Dict[str, str]]] = None

  def __iter__(self) -> 'Query':
    return self

  def __next__(self) -> T_Model:
    if not self._paginator:
      self._begin()
      assert self._paginator is not None
    while True:
      if self._page_items is None:
        self._page_items = iter(next(self._paginator).results)
      try:
        row = next(self._page_items)
        break
      except StopIteration:
        self._page_items = None
        continue

    return self._db._build_model_from_row(self._model, row)

  def _begin(self) -> None:
    self._paginator = self._db._client.paginated_database_table_rows(
      self._mapping.table_id,
      filter=self._filters)

  @property
  def executed(self) -> bool:
    """
    Returns #True if the query has been executed.
    """

    return self._paginator is not None

  def filter(self, *filters: Filter) -> 'Query[T_Model]':
    """
    Filter the query with the specified filters. This method modified the query.
    """

    if self.executed:
      raise RuntimeError('Query has already been executed')
    for filter in filters:
      self._filters.append(self._db._preprocess_filter(filter))
    return self

  def page_size(self, page_size: int) -> 'Query[T_Model]':
    """
    Set the number of rows to return per call to the Baserow API. This method modified the query.
    """

    if self.executed:
      raise RuntimeError('Query has already been executed')
    self._page_size = page_size
    return self

  def first(self) -> T_Model:
    """
    Return the first row from the query. This method executes the query.

    If the query returns no rows, a #NoRowReturned exception will be raised instead.
    """

    if self.executed:
      raise RuntimeError('Query has already been executed')

    self.page_size(1)
    try:
      return next(self)
    except StopIteration:
      raise NoRowReturned




from .column import Column, ForeignKey
from .database import Database
from .exc import BaserowOrmException, NoRowReturned
from .mapping import DatabaseMapping, generate_mapping
from .model import Model

__all__ = [
  'Column',
  'ForeignKey',
  'Database',
  'BaserowOrmException',
  'NoRowReturned',
  'DatabaseMapping',
  'generate_mapping',
  'Model'
]




import dataclasses
import enum
import typing as t

from databind.core.annotations import union

T = t.TypeVar('T')


class Permissions(enum.Enum):
  MEMBER = enum.auto()
  ADMIN = enum.auto()


@dataclasses.dataclass
class User:
  id: int
  first_name: str
  username: str
  is_staff: bool


@dataclasses.dataclass
class Group:
  id: int
  name: str


@dataclasses.dataclass
class OrderedGroup(Group):
  order: int


@dataclasses.dataclass
class PermissionedOrderedGroup(OrderedGroup):
  permissions: Permissions


@dataclasses.dataclass
class Table:
  id: int
  name: str
  order: int
  database_id: int


@union(style=union.Style.flat, constructible=True)
@dataclasses.dataclass
class TableField:
  id: int
  table_id: int
  name: str
  order: int
  primary: bool


@dataclasses.dataclass
class Application:
  id: int
  name: str
  order: int
  type: str
  group: Group
  tables: t.List[Table]


@dataclasses.dataclass
class Page(t.Generic[T]):
  count: int
  previous: t.Optional[int]
  next: t.Optional[int]
  results: t.List[T]


from . import field_types




import dataclasses
import json
import logging
import typing as t
from pathlib import Path

import databind.json
import requests

from .filter import Filter, FilterType
from .types import Application, Page, PermissionedOrderedGroup, Table, TableField, User

log = logging.getLogger(__name__)
DEFAULT_CREDENTIALS_FILE = '.baserow-creds.json'


@dataclasses.dataclass
class ApiError(Exception):
  error: str
  detail: str

  def __str__(self) -> str:
    return f'{self.error}: {self.detail}'


class BaseClient:
  """
  Base class for the Baserow client which handles the authentication and request session.

  A JWT is needed when performing requests that are scoped to user interactions (e.g. operations that
  are expected to be run through the UI). A token should be used if only a subset of the Baserow API
  is used to create/read/write/delete rows.
  """

  def __init__(self, url: str, token: t.Optional[str] = None, jwt: t.Optional[str] = None) -> None:
    if token and jwt:
      raise ValueError(f'token/jwt can not be specified at the same time')

    self._url = url.rstrip('/')
    self._session = requests.Session()
    self._jwt: t.Optional[str] = None
    self._token: t.Optional[str] = None

    if jwt:
      self.jwt = jwt
    elif token:
      self.token = token

  def _request(self, method: str, path: str, **kwargs) -> requests.Response:
    response = self._session.request(method, self._url + '/' + path.lstrip('/'), **kwargs)
    if (response.status_code // 100) in (4, 5) and response.headers.get('Content-Type') == 'application/json':
      data = response.json()
      log.debug('Error from %s %s: %s', method, path, data)
      raise ApiError(data.get('error', 'UNKNOWN'), data.get('detail', '???'))
    response.raise_for_status()
    return response

  @property
  def jwt(self) -> t.Optional[str]:
    return self._jwt

  @jwt.setter
  def jwt(self, jwt: str) -> None:
    self._jwt = jwt
    self._token = None
    self._session.headers['Authorization'] = f'JWT {jwt}'

  @property
  def token(self) -> t.Optional[str]:
    return self._token

  @token.setter
  def token(self, token: str) -> None:
    self._jwt = None
    self._token = token
    self._session.headers['Authorization'] = f'Token {token}'


class BaserowClient(BaseClient):
  """
  Client for Baserow servers.
  """

  def get_settings(self) -> t.Dict[str, t.Any]:
    return self._request('GET', '/api/settings/').json()

  def update_settings(self, settings: t.Dict[str, t.Any]) -> None:
    self._request('PATCH', '/api/settings/update/', json=settings)

  def token_auth(self, username: str, password: str) -> t.Tuple[User, str]:
    payload = {'username': username, 'password': password}
    response = self._request('POST', '/api/user/token-auth/', json=payload).json()
    return databind.json.load(response['user'], User), response['token']

  def token_refresh(self, token: str) -> t.Tuple[User, str]:
    payload = {'token': token}
    response = self._request('POST', '/api/user/token-refresh/', json=payload).json()
    return databind.json.load(response['user'], User), response['token']

  def create_user(
    self,
    name: str,
    email: str,
    password: str,
    authenticate: bool = False,
    group_invitation_token: t.Optional[str] = None,
    template_id: t.Optional[int] = None
  ) -> t.Tuple[User, t.Optional[str]]:

    payload: t.Dict[str, t.Union[str, bool, int]] = {'name': name, 'email': email, 'password': password}
    if authenticate:
      payload['authenticate'] = authenticate
    if group_invitation_token:
      payload['group_invitation_token'] = group_invitation_token
    if template_id:
      payload['template_id'] = template_id

    response = self._request('POST', '/api/user/', json=payload).json()
    return databind.json.load(response['user'], User), response.get('token')

  def list_groups(self) -> t.List[PermissionedOrderedGroup]:
    response = self._request('GET', '/api/groups/').json()
    return databind.json.load(response, t.List[PermissionedOrderedGroup])

  def create_group(self, name: str) -> PermissionedOrderedGroup:
    response = self._request('POST', '/api/groups/', json={'name': name}).json()
    return databind.json.load(response, PermissionedOrderedGroup)

  def list_all_applications(self) -> t.List[Application]:
    response = self._request('GET', '/api/applications/').json()
    return databind.json.load(response, t.List[Application])

  def get_database_table(self, table_id: int) -> Table:
    response = self._request('GET', f'/api/database/tables/{table_id}').json()
    return databind.json.load(response, Table)

  def update_database_table(self, table_id: int, name: str) -> Table:
    response = self._request('PATCH', f'/api/database/tables/{table_id}', json={'name': name}).json()
    return databind.json.load(response, Table)

  def list_database_tables(self, database_id: int) -> t.List[Table]:
    response = self._request('GET', f'/api/database/tables/database/{database_id}').json()
    return databind.json.load(response, t.List[Table])

  def list_database_table_fields(self, table_id: int) -> t.List[TableField]:
    response = self._request('GET', f'/api/database/fields/table/{table_id}').json()
    return databind.json.load(response, t.List[TableField])

  def list_database_table_rows(
    self,
    table_id: int,
    exclude: t.Optional[t.List[str]] = None,
    filter: t.Optional[t.List[Filter]] = None,
    filter_type: t.Optional[FilterType] = None,
    include: t.Optional[t.List[str]] = None,
    order_by: t.Optional[t.List[str]] = None,
    page: t.Optional[int] = None,
    search: t.Optional[str] = None,
    size: t.Optional[int] = None,
    user_field_names: bool = False,
  ) -> Page[t.Dict[str, t.Any]]:

    params: t.Dict[str, t.Optional[str]] = {}
    if exclude is not None:
      params['exclude'] = ','.join(exclude)
    if filter is not None:
      params.update(dict(f.to_query_parameter() for f in filter))
    if filter_type is not None:
      params['filter_type'] = filter_type.name
    if include is not None:
      params['include'] = ','.join(include)
    if order_by is not None:
      params['order_by'] = ','.join(order_by)
    if page is not None:
      params['page'] = str(page)
    if search is not None:
      params['search'] = search
    if size is not None:
      params['size'] = str(size)
    if user_field_names:
      params['user_field_names'] = 'True'

    response = self._request('GET', f'/api/database/rows/table/{table_id}/', params=params).json()
    if page is None:
      page = 1

    return Page(
      response['count'],
      page - 1 if page > 1 else None,
      page + 1 if response['next'] else None,
      response['results'])

  def create_database_table_row(self, table_id: int, record: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
    return self._request('POST', f'/api/database/rows/table/{table_id}/', json=record).json()

  def update_database_table_row(self, table_id: int, row_id: int, record: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
    return self._request('PATCH', f'/api/database/rows/table/{table_id}/{row_id}/', json=record).json()

  def get_database_table_row(self, table_id: int, row_id: int) -> t.Dict[str, t.Any]:
    return self._request('GET', f'/api/database/rows/table/{table_id}/{row_id}/').json()

  # Extra

  def login(self, username: str, password: str, cache: t.Union[bool, str] = False) -> User:
    """
    A convenience method to log into Baserow using the specified *username* and *password* and updating
    the current client object. If *cache* is enabled or is a filename, it will be used to load a cached
    token for the Baserow URL and username combination to reuse a previously generated JWT. If the reused
    JWT is not valid anymore, the credentials will be used to generate a new one.
    """

    cache_fn = cache if isinstance(cache, str) else None
    if cache:
      user = self.load(username, cache_fn)
      if user:
        return user

    log.info('Creating new JWT')
    user, self.jwt = self.token_auth(username, password)
    if cache:
      self.save(username, cache_fn)
    return user

  def load(self, username: str, filename: t.Optional[str] = None, raise_: bool = False, refresh: bool = True) -> t.Optional[User]:
    """
    Loads an existing JWT from the given *filename* or the #DEFAULT_CREDENTIALS_FILE. Returns #True if a token
    was loaded, #False otherwise. If a token is loaded, it will be immediately refreshed.
    """

    path = Path(filename or DEFAULT_CREDENTIALS_FILE)
    if not path.exists():
      if raise_:
        raise FileNotFoundError(path)
      return None

    try:
      data = json.loads(path.read_text())
    except json.JSONDecodeError:
      log.error('Unable to parse JSON file %s', path)
      if raise_:
        raise
      return None

    jwt = data.get(self._url, {}).get(username)
    if jwt:
      log.info('Refreshing JWT')
      user, self.jwt = self.token_refresh(jwt)
      self.save(username, filename)
      return user

    return None

  def save(self, username: str, filename: t.Optional[str] = None) -> None:
    """
    Saves the JWt of the client into *filename* or the given #DEFAULT_CREDENTIALS_FILE.
    """

    if not self.jwt:
      raise ValueError(f'No JWT set')

    path = Path(filename or DEFAULT_CREDENTIALS_FILE)
    data = json.loads(path.read_text()) if path.exists() else {}
    data.setdefault(self._url, {})[username] = self.jwt
    path.write_text(json.dumps(data, indent=2))

  def paginated_database_table_rows(
    self,
    table_id: int,
    exclude: t.Optional[t.List[str]] = None,
    filter: t.Optional[t.List[Filter]] = None,
    filter_type: t.Optional[FilterType] = None,
    include: t.Optional[t.List[str]] = None,
    order_by: t.Optional[t.List[str]] = None,
    search: t.Optional[str] = None,
    size: t.Optional[int] = None,
    user_field_names: bool = False,
  ) -> t.Generator[Page[t.Dict[str, t.Any]], None, None]:

    page_number = None
    while True:
      page = self.list_database_table_rows(
        table_id=table_id,
        exclude=exclude,
        filter=filter,
        filter_type=filter_type,
        include=include,
        order_by=order_by,
        page=page_number,
        search=search,
        size=size,
        user_field_names=user_field_names,
      )
      if page.results:
        yield page
      if not page.next:
        break
      page_number = page.next




import dataclasses
import enum
import typing as t

from databind.core.annotations import union

from .types import TableField


class NumberType(enum.Enum):
  INTEGER = enum.auto()
  DECIMAL = enum.auto()


@dataclasses.dataclass
class SelectOption:
  id: int
  value: str
  color: str


@union.subtype(TableField, 'text')
@dataclasses.dataclass
class TextTableField(TableField):
  text_default: str


@union.subtype(TableField, 'long_text')
@dataclasses.dataclass
class LongTextTableField(TableField): pass


@union.subtype(TableField, 'number')
@dataclasses.dataclass
class NumberTableField(TableField):
  number_decimal_places: int
  number_negative: bool
  number_type: NumberType


@union.subtype(TableField, 'single_select')
@dataclasses.dataclass
class SingleSelectTableField(TableField):
  select_options: t.List[SelectOption]


@union.subtype(TableField, 'url')
@dataclasses.dataclass
class UrlTableField(TableField):
  pass


@union.subtype(TableField, 'link_row')
@dataclasses.dataclass
class LinkRowTableField(TableField):
  link_row_table: int
  link_row_related_field: int


@union.subtype(TableField, 'boolean')
@dataclasses.dataclass
class BooleanTableField(TableField): pass


@union.subtype(TableField, 'file')
@dataclasses.dataclass
class FileTableField(TableField): pass




import dataclasses
import datetime
import enum
import typing as t

Orderable = t.Union[int, float]
BasicType = t.Union[bool, str, Orderable]
Date = t.Union[datetime.date, datetime.datetime]
ValueType = t.Union[BasicType, Date]


class FilterType(enum.Enum):
  OR = enum.auto()
  AND = enum.auto()


class FilterMode(enum.Enum):
  equal = enum.auto()
  not_equal = enum.auto()
  filename_contains = enum.auto()
  contains = enum.auto()
  contains_not = enum.auto()
  higher_than = enum.auto()
  lower_than = enum.auto()
  date_equal = enum.auto()
  date_before = enum.auto()
  date_after = enum.auto()
  date_not_equal = enum.auto()
  date_equals_today = enum.auto()
  date_equals_month = enum.auto()
  date_equals_year = enum.auto()
  single_select_equal = enum.auto()
  single_select_not_equal = enum.auto()
  link_row_has = enum.auto()
  link_row_has_not = enum.auto()
  boolean = enum.auto()
  empty = enum.auto()
  not_empty = enum.auto()


@dataclasses.dataclass
class Filter:
  field: str
  filter: FilterMode
  value: t.Optional[ValueType]

  def to_query_parameter(self) -> t.Tuple[str, t.Optional[str]]:
    key = f'filter__{self.field}__{self.filter.name}'
    value: t.Optional[str]
    if isinstance(self.value, datetime.datetime):
      value = self.value.strftime('%Y-%m-%dT%H:%M:%S%z')
    elif isinstance(self.value, datetime.date):
      value = self.value.strftime('%Y-%m-%d')
    elif self.value is None:
      value = None
    else:
      value = str(self.value)
    return (key, value)


class Column:
  """
  A helper class to build #Filter#s.
  """

  def __init__(self, name: str) -> None:
    self._name = name

  def equal(self, value: ValueType) -> Filter:
    return Filter(self._name, FilterMode.equal, value)

  def not_equal(self, value: ValueType) -> Filter:
    return Filter(self._name, FilterMode.not_equal, value)

  def filename_contains(self, value: str) -> Filter:
    return Filter(self._name, FilterMode.filename_contains, value)

  def contains(self, value: str) -> Filter:
    return Filter(self._name, FilterMode.contains, value)

  def contains_not(self, value: str) -> Filter:
    return Filter(self._name, FilterMode.contains_not, value)

  def higher_than(self, value: Orderable) -> Filter:
    return Filter(self._name, FilterMode.higher_than, value)

  def lower_than(self, value: Orderable) -> Filter:
    return Filter(self._name, FilterMode.lower_than, value)

  def date_equal(self, value: Date) -> Filter:
    return Filter(self._name, FilterMode.date_equal, value)

  def date_before(self, value: Date) -> Filter:
    return Filter(self._name, FilterMode.date_before, value)

  def date_after(self, value: Date) -> Filter:
    return Filter(self._name, FilterMode.date_after, value)

  def date_not_equal(self, value: Date) -> Filter:
    return Filter(self._name, FilterMode.date_not_equal, value)

  def date_equals_today(self) -> Filter:
    return Filter(self._name, FilterMode.date_equals_today, None)

  def date_equals_month(self, month: int) -> Filter:
    return Filter(self._name, FilterMode.date_equals_month, month)

  def date_equals_year(self, year: int) -> Filter:
    return Filter(self._name, FilterMode.date_equals_year, year)

  def single_select_equal(self, value: str) -> Filter:
    return Filter(self._name, FilterMode.single_select_equal, value)

  def single_select_not_equal(self, value: str) -> Filter:
    return Filter(self._name, FilterMode.single_select_not_equal, value)

  def link_row_has(self, value: int) -> Filter:
    return Filter(self._name, FilterMode.link_row_has, value)

  def link_row_has_not(self, value: int) -> Filter:
    return Filter(self._name, FilterMode.link_row_has_not, value)

  # def boolean(self, value: ValueType) -> Filter:
  #   return Filter(self._name, FilterMode.boolean, value)

  def empty(self) -> Filter:
    return Filter(self._name, FilterMode.empty, None)

  def not_empty(self) -> Filter:
    return Filter(self._name, FilterMode.not_empty, None)




import dataclasses
import json
import typing as t
from pathlib import Path

import databind.json

from ..client import BaserowClient
from .model import Model, ModelMappingDescription


@dataclasses.dataclass
class ModelMapping:
  """
  Contains the mapping details for a model.
  """

  #: The table ID in Baserow.
  table_id: int

  #: Maps model attribute names defined to internal field IDs.
  fields: t.Dict[str, int]

  def __post_init__(self) -> None:
    self._reverse_fields: t.Optional[t.Dict[int, str]] = {}

  @property
  def reverse_fields(self) -> t.Dict[int, str]:
    if self._reverse_fields is None or len(self.fields) != len(self._reverse_fields):
      self._reverse_fields = {v: k for k, v in self.fields.items()}
    return self._reverse_fields


@dataclasses.dataclass
class DatabaseMapping:
  """
  Contains the mapping details for a database and its models.
  """

  #: The database ID in Baserow.
  database_id: int

  #: The models included in this database mapping.
  models: t.Dict[str, ModelMapping]

  def to_json(self) -> t.Dict[str, t.Any]:
    return databind.json.dump(self)  # type: ignore

  @staticmethod
  def from_json(data: t.Dict[str, t.Any]) -> 'DatabaseMapping':
    return databind.json.load(data, DatabaseMapping)

  def save(self, filename: t.Union[str, Path], indent: t.Optional[int] = None) -> None:
    Path(filename).write_text(json.dumps(self.to_json(), indent=indent))

  @staticmethod
  def load(filename: t.Union[str, Path]) -> 'DatabaseMapping':
    return databind.json.loads(Path(filename).read_text(), DatabaseMapping)


def generate_mapping(
  client: BaserowClient,
  dbname: str,
  *models: t.Union[ModelMappingDescription, t.Type[Model]],
) -> DatabaseMapping:
  """
  Generate a database mapping from the specified database name and models. A #ModelMappingDescription can
  be specified instead of a #Model class either if the #Model.__tablename__ is not defined or if custom
  field names need to be overriden for this mapping.

  # Example

  ```py
  from baserow.orm import generate_mapping, DatabaseMapping

  if client.jwt:
    mapping = generate_mapping(client, 'Blog', Post.of('Posts', {'name': 'Product Name'}))
    mapping.save('mapping.json')
  else:
    mapping = DatabaseMapping.load('mapping.json')

  db = Database(client, mapping)
  ```
  """

  try:
    db = next(db for db in client.list_all_applications() if db.name == dbname)
  except StopIteration:
    raise ValueError(f'database {dbname!r} does not exist')

  model_mappings: t.Dict[str, ModelMapping] = {}

  for model in models:
    if isinstance(model, type):
      assert issubclass(model, Model), model
      if not model.__tablename__:
        raise ValueError(f'Missing __tablename__ for model {model.__id__!r}')
      model = ModelMappingDescription(model.__id__, model.__columns__, model.__tablename__, {})
    try:
      table = next(t for t in db.tables if t.name == model.table_name)
    except StopIteration:
      raise ValueError(f'table {dbname!r}/{model.table_name!r} does not exist')

    table_fields = {f.name: f for f in client.list_database_table_fields(table.id)}
    fields: t.Dict[str, int] = {}
    for key, column in model.columns.items():
      name = model.field_name_overrides.get(key) or column.name
      if name not in table_fields:
        raise ValueError(f'field {dbname!r}/{model.table_name!r}/{name} does not exist')
      fields[key] = table_fields[name].id

    model_mappings[model.model_id] = ModelMapping(table.id, fields)

  return DatabaseMapping(db.id, model_mappings)




import argparse
import getpass
import importlib
import json
import typing as t

from baserow.orm.model import Model

from ..client import BaserowClient
from .mapping import DatabaseMapping, ModelMappingDescription, generate_mapping

parser = argparse.ArgumentParser()
parser.add_argument('dbname', help='The database name to use.')
parser.add_argument('models', nargs='+', help='One or more absolute model IDs and their Baserow table name separated by colons.')
parser.add_argument('--url', required=True, help='The URL to your Baserow instance.')
parser.add_argument('--user', help='The username to authenticate with.')
parser.add_argument('--password', help='The password for the given --user. If omitted, the password will be asked.')
parser.add_argument('--jwt', help='A JWT to use instead of generating a new one with --user and --password.')
parser.add_argument('--write-to', help='The path to write the generated mappings to. If not specified, will be stdout.')


def main():
  args = parser.parse_args()
  client = BaserowClient(args.url)

  if args.user:
    if not args.password:
      args.password = getpass.getpass(f'Password for {args.user}: ')
    client.login(args.user, args.password)
  elif args.jwt:
    client.jwt = args.jwt
  else:
    parser.error('need at least --user or --jwt')

  models: t.List[t.Union[t.Type[Model], ModelMappingDescription]] = []
  for spec in args.models:
    if ':' in spec:
      model_id, table_name = spec.split(':')
    else:
      model_id = spec
      table_name = None
    module_name, class_name = model_id.rpartition('.')[::2]
    module = importlib.import_module(module_name)
    model: t.Type[Model] = getattr(module, class_name)
    if table_name:
      models.append(ModelMappingDescription(model_id, model.__columns__, table_name, {}))
    else:
      models.append(model)

  mapping = generate_mapping(client, args.dbname, *models)
  if args.write_to:
    mapping.save(args.write_to)
  else:
    print(json.dumps(mapping.to_json(), indent=2))


if __name__ == '__main__':
  main()
